# This code is extracted from the tornado project -- they did some nice
# pretty logging. See: https://tornadoweb.org for original project and
# source code.
#
# Copyright 2012 Facebook
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import curses
import logging
import sys
import typing

import colorama


@typing.overload
def to_unicode(value: str) -> str:
	pass


@typing.overload  # noqa: F811
def to_unicode(value: bytes) -> str:
	pass


@typing.overload  # noqa: F811
def to_unicode(value: None) -> None:
	pass


bytes_type = bytes
unicode_type = str
basestring_type = str

_TO_UNICODE_TYPES = (unicode_type, type(None))


def to_unicode(value: typing.Union[None, str, bytes]) -> typing.Optional[str]:  # noqa: F811
	"""Converts a string argument to a unicode string.

	If the argument is already a unicode string or None, it is returned
	unchanged.  Otherwise it must be a byte string and is decoded as utf8.
	"""
	if isinstance(value, _TO_UNICODE_TYPES):
		return value
	if not isinstance(value, bytes):
		raise TypeError("Expected bytes, unicode, or None; got %r" % type(value))
	return value.decode("utf-8")


# to_unicode was previously named _unicode not because it was private,
# but to avoid conflicts with the built-in unicode() function/type
_unicode = to_unicode


def _safe_unicode(s: typing.Any) -> str:
	try:
		return _unicode(s)
	except UnicodeDecodeError:
		return repr(s)


def _stderr_supports_color() -> bool:
	try:
		if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
			if curses:
				curses.setupterm()
				if curses.tigetnum("colors") > 0:
					return True
			elif colorama:
				if sys.stderr is getattr(
						colorama.initialise, "wrapped_stderr", object()
				):
					return True
	except Exception:
		# Very broad exception handling because it's always better to
		# fall back to non-colored logs than to break at startup.
		pass
	return False


class TornadoPrettyLogFormatter(logging.Formatter):
	"""Log formatter used in Tornado.

	Key features of this formatter are:

	* Color support when logging to a terminal that supports it.
	* Timestamps on every log line.
	* Robust against str/bytes encoding problems.

	This formatter is enabled automatically by
	`tornado.options.parse_command_line` or `tornado.options.parse_config_file`
	(unless ``--logging=none`` is used).

	Color support on Windows versions that do not support ANSI color codes is
	enabled by use of the colorama__ library. Applications that wish to use
	this must first initialize colorama with a call to ``colorama.init``.
	See the colorama documentation for details.

	__ https://pypi.python.org/pypi/colorama

	.. versionchanged:: 4.5
	   Added support for ``colorama``. Changed the constructor
	   signature to be compatible with `logging.config.dictConfig`.
	"""

	#DEFAULT_FORMAT = "%(color)s[%(levelname)5.5s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s"  # noqa: E501
	DEFAULT_FORMAT = "%(color)s[%(levelname)s %(asctime)s]%(end_color)s %(message)s"  # noqa: E501
	DEFAULT_DATE_FORMAT = "%H:%M:%S"
	DEFAULT_COLORS = {
		logging.DEBUG: 4,  # Blue
		logging.INFO: 2,  # Green
		logging.WARNING: 3,  # Yellow
		logging.ERROR: 1,  # Red
		logging.CRITICAL: 5,  # Magenta
	}

	def __init__(
			self,
			fmt: str = DEFAULT_FORMAT,
			datefmt: str = DEFAULT_DATE_FORMAT,
			style: str = "%",
			color: bool = True,
			colors: typing.Dict[int, int] = DEFAULT_COLORS,
	) -> None:
		r"""
		:arg bool color: Enables color support.
		:arg str fmt: Log message format.
		  It will be applied to the attributes dict of log records. The
		  text between ``%(color)s`` and ``%(end_color)s`` will be colored
		  depending on the level if color support is on.
		:arg dict colors: color mappings from logging level to terminal color
		  code
		:arg str datefmt: Datetime format.
		  Used for formatting ``(asctime)`` placeholder in ``prefix_fmt``.

		.. versionchanged:: 3.2

		   Added ``fmt`` and ``datefmt`` arguments.
		"""
		logging.Formatter.__init__(self, datefmt=datefmt)
		self._fmt = fmt

		self._colors = {}  # type: Dict[int, str]
		if color and _stderr_supports_color():
			if curses is not None:
				fg_color = curses.tigetstr("setaf") or curses.tigetstr("setf") or b""

				for levelno, code in colors.items():
					# Convert the terminal control characters from
					# bytes to unicode strings for easier use with the
					# logging module.
					self._colors[levelno] = unicode_type(
						curses.tparm(fg_color, code), "ascii"
					)
				self._normal = unicode_type(curses.tigetstr("sgr0"), "ascii")
			else:
				# If curses is not present (currently we'll only get here for
				# colorama on windows), assume hard-coded ANSI color codes.
				for levelno, code in colors.items():
					self._colors[levelno] = "\033[2;3%dm" % code
				self._normal = "\033[0m"
		else:
			self._normal = ""

	def format(self, record: typing.Any) -> str:
		try:
			message = record.getMessage()
			assert isinstance(message, basestring_type)  # guaranteed by logging
			# Encoding notes:  The logging module prefers to work with character
			# strings, but only enforces that log messages are instances of
			# basestring.  In python 2, non-ascii bytestrings will make
			# their way through the logging framework until they blow up with
			# an unhelpful decoding error (with this formatter it happens
			# when we attach the prefix, but there are other opportunities for
			# exceptions further along in the framework).
			#
			# If a byte string makes it this far, convert it to unicode to
			# ensure it will make it out to the logs.  Use repr() as a fallback
			# to ensure that all byte strings can be converted successfully,
			# but don't do it by default so we don't add extra quotes to ascii
			# bytestrings.  This is a bit of a hacky place to do this, but
			# it's worth it since the encoding errors that would otherwise
			# result are so useless (and tornado is fond of using utf8-encoded
			# byte strings wherever possible).
			record.message = _safe_unicode(message)
		except Exception as e:
			record.message = "Bad message (%r): %r" % (e, record.__dict__)

		record.asctime = self.formatTime(record, typing.cast(str, self.datefmt))

		if record.levelno in self._colors:
			record.color = self._colors[record.levelno]
			record.end_color = self._normal
		else:
			record.color = record.end_color = ""

		formatted = self._fmt % record.__dict__

		if record.exc_info:
			if not record.exc_text:
				record.exc_text = self.formatException(record.exc_info)
		if record.exc_text:
			# exc_text contains multiple lines.  We need to _safe_unicode
			# each line separately so that non-utf8 bytes don't cause
			# all the newlines to turn into '\n'.
			lines = [formatted.rstrip()]
			lines.extend(_safe_unicode(ln) for ln in record.exc_text.split("\n"))
			formatted = "\n".join(lines)
		return formatted.replace("\n", "\n    ")




import os
import subprocess
import jinja2

import dyne.org.funtoo.boxer.containers as containers
from subpop.util import AttrDict


dockerfile = """
# we are going to use alpine as our bootstrap container
ARG BOOTSTRAP
FROM ${BOOTSTRAP:-alpine:3.13} as builder

WORKDIR /funtoo

ARG STAGE3="{{stage}}"

COPY "${STAGE3}" /
RUN echo "Building Funtoo Container image." \
 && sleep 3 \
 && apk --no-cache add gnupg tar wget xz \
 && tar xpf /${STAGE3} --xattrs --xattrs-include='*' --numeric-owner \
 && sed -i -e 's/#rc_sys=""/rc_sys="docker"/g' etc/rc.conf \
 && rm -f etc/localtime \
 && ln -s ../usr/share/zoneinfo/UTC etc/localtime \
 && rm /*.xz

FROM scratch

WORKDIR /
COPY --from=builder /funtoo/ /
RUN echo "Customizing funtoo for the needs of the people." \
 && rm /etc/fstab \
 && touch /etc/fstab \
 && sed -i -e 's/^c/#c/g' /etc/inittab \
 && rm -rf /usr/src/linux* \
 && rm -rf /var/db/pkg/sys-kernel/debian-sources* \
 && rc-update del sysctl boot \
 && rc-update del hostname boot \
 && rc-update del loopback boot \
 && rc-update del udev sysinit \
 && rc-update del bootmisc boot \
 && rm -rf /var/git/* \
 && rm -rf /lib/modules/*
ENTRYPOINT ["/sbin/init"]
"""

def get_docker_args():
	return AttrDict({
		"push": containers.model.push,
		"tag": containers.model.tag,
		"stage": containers.model.stage.name,
	})


def cmd(cwd, cmd: list, desc: str = None):
	if desc:
		containers.model.log.info(desc)
	containers.model.log.debug(f"cwd: {cwd}")
	containers.model.log.info(f"Running command: {cmd}")
	result = subprocess.run(cmd, cwd=cwd)
	if result.returncode != 0:
		containers.model.log.error(f"Error: exit code {result.returncode}.")
		raise ChildProcessError()


def create_container():
	args = get_docker_args()
	template = jinja2.Template(dockerfile)
	outfile = os.path.join(containers.model.tmp, 'Dockerfile')
	with open(outfile, "wb") as myf:
		myf.write(template.render(**args).encode("utf-8"))
		containers.model.log.debug(f"Output written to {outfile}.")
	dest_stage = os.path.join(containers.model.tmp, containers.model.stage.name)
	if os.path.exists(dest_stage):
		os.unlink(dest_stage)
	cmd(containers.model.tmp, ["cp", containers.model.stage, "."], desc="Copy stage to temp directory")
	base_cmd = ["docker", "build", "."]
	if args.tag:
		base_cmd += ["-t", args.tag]
	cmd(containers.model.tmp, base_cmd, desc="Creating docker container")
	if args.tag and args.push:
		cmd(containers.model.tmp, ["docker", "push", args.tag], desc="Pushing container")
	return True




import argparse
import subprocess

from boxer.model import get_model

model = get_model('containers')


def cmd(cwd, cmd: list, desc: str = None):
	if desc:
		model.log.info(desc)
	model.log.debug(f"cwd: {cwd}")
	model.log.info(f"Running command: {cmd}")
	result = subprocess.run(cmd, cwd=cwd)
	if result.returncode != 0:
		model.log.error(f"Error: exit code {result.returncode}.")
		raise ChildProcessError()


def parse_args(cli_config, args=None, parse_known=False):
	ap = argparse.ArgumentParser()
	for arg, kwargs in cli_config.items():
		if "positional" in kwargs and kwargs["positional"]:
			new_kwargs = kwargs.copy()
			del new_kwargs["positional"]
			ap.add_argument(arg, **new_kwargs)
		else:
			if "os" in kwargs:
				del kwargs["os"]
			ap.add_argument("--" + arg, **kwargs)
	if parse_known:
		return ap.parse_known_args(args)
	else:
		return ap.parse_args(args)



#!/usr/bin/env python3

import os

from subpop.config import ConfigurationError


class Locator:
	start_path = None
	root: str = None
	expected_files = []
	"""
	This method will look, from the current directory, and find the 'context' of where
	we are in a kit-fixups repository, so we know where the main kit-fixups repository
	is and what is our current working set for autogeneration.
	"""

	def found_root(self, cur_path) -> bool:
		for file in self.expected_files:
			if not os.path.exists(os.path.join(cur_path, file)):
				return False
		return True

	def find_root(self):
		"""
		This method starts from ``start_path`` and looks backwards in the path structure until it
		gets to the point where it finds a 'profiles/repo_name' and/or 'metadata/layout.conf' file,
		which indicates that it is at the root of an overlay, and then it returns this value.

		If it gets to / without finding these files, it returns None as a failure indicator that
		it couldn't find the overlay that we are currently inside of via the current working
		directory.
		"""
		cur_path = self.start_path
		while (
				cur_path != "/"
				and not self.found_root(cur_path)
		):
			cur_path = os.path.dirname(cur_path)
		if cur_path == "/":
			cur_path = None
		return cur_path

	def __init__(self, start_path=None):
		self.start_path = start_path if start_path else os.getcwd()
		self.root = self.find_root()
		if self.root is None:
			raise ConfigurationError(f"Could not determine context in {self.start_path}. Trying to find these marker files: {self.expected_files}")

class GitRepositoryLocator(Locator):
	expected_files = [".git"]



import os
import shutil
import sys

import dyne.org.funtoo.boxer.containers as containers

from boxer.config.base import SingularityConfig, DockerConfig


async def start():
	errcode = 0
	try:
		if isinstance(containers.model, SingularityConfig):
			success = containers.singularity.create_container()
		elif isinstance(containers.model, DockerConfig):
			success = containers.docker.create_container()
	except ChildProcessError as e:
		errcode = 1
	finally:
		if os.path.exists(containers.model.tmp):
			shutil.rmtree(containers.model.tmp)
	if not success:
		errcode = 1
	sys.exit(errcode)




#!/usr/bin/env python3

# This was originally added so that support classes can get access to the model without having to be
# part of the subsystem. The original way to use it was to have the MergeConfig initialize call
# set_model(self) and then a support library could call get_model("metatools.merge"). Now I am trying
# to extend this so that get_model() by itself will return *any* model, and get_model("metatools")
# will return any model that is part of metatools. This can be used with utility libraries that
# need functionality that are in the metatools BaseConfig, so *any* metatools config object will
# do.

MODELS = {}


class ModelWrapper:

	def __init__(self, name=None):
		self.name = name

	def __getattr__(self, item):
		try:
			return getattr(MODELS[self.name], item)
		except KeyError as ke:
			print(f"Available models: {MODELS.keys()}")
			raise ke


def set_model(name, model):
	model_split = name.split('.')
	if len(model_split) == 1:
		MODELS[name] = model
	elif len(model_split) == 2:
		MODELS[name] = model
		MODELS[model_split[0]] = model
	# Set default model to last set:
	MODELS[None] = model


def get_model(name):
	if name is None:
		return ModelWrapper()
	else:
		return ModelWrapper(name)




import os

import dyne.org.funtoo.boxer.containers as containers

from boxer.common import cmd


def remove(rootfs, path):
	path = path.lstrip("/")
	cmd(rootfs, ["rm", "-rf", path], desc=f"Cleaning rootfs: {path}")


def create_container():
	if containers.model.out is None:
		out = os.path.join(os.getcwd(), "funtoo-stage3.sif")
	else:
		out = containers.model.out
	if os.path.exists(out):
		if not containers.model.force:
			containers.model.log.error(f"{out} exists -- use --force to overwrite. Aborting.")
			return False
	rootfs = os.path.join(containers.model.tmp, "rootfs")
	os.makedirs(rootfs, exist_ok=False)
	cmd(rootfs,
	    ["tar", "--numeric-owner", "--xattrs", "--xattrs-include='*'", "-xpf", str(containers.model.stage), "."],
	    desc="Extracting stage tarball")
	remove(rootfs, "/usr/src/linux*")
	remove(rootfs, "/lib/modules/*")
	remove(rootfs, "/var/git/*")
	remove(rootfs, "/var/db/pkg/sys-kernel/debian-sources*")
	d_dir = os.path.join(rootfs, ".singularity.d")
	os.makedirs(os.path.join(d_dir), exist_ok=True)
	if os.path.exists(os.path.join(rootfs, "etc/mtab")):
		os.unlink(os.path.join(rootfs, "etc/mtab"))
	with open(os.path.join(rootfs, "etc/fstab"), "w") as f:
		f.write("proc /proc proc defaults 0 0\n")
	with open(os.path.join(rootfs, "etc/rc.conf"), "a") as f:
		f.write("rc_sys=lxd\n")
	with open(os.path.join(d_dir, "runscript"), "w") as f:
		f.write("""#!/bin/bash
exec /bin/bash --login
""")
	build_cmd = ["singularity", "build" ]
	if containers.model.force:
		build_cmd.append("-F")
	build_cmd += [ out, rootfs ]
	cmd(containers.model.tmp, build_cmd, desc="Creating singularity container")


import setuptools
from subpop.pkg import Packager, SubPopSetupInstall

pkgr = Packager()

with open("README.rst", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="funtoo-boxer",
	version="1.0.3",
	author="Daniel Robbins",
	author_email="drobbins@funtoo.org",
	description="Funtoo framework for container generation.",
	long_description=long_description,
	long_description_content_type="text/x-rst",
	url="https://code.funtoo.org/bitbucket/users/drobbins/repos/funtoo-boxer/browse",
	scripts=["bin/boxer"],
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: Apache Software License",
		"Operating System :: POSIX :: Linux",
	],
	python_requires=">=3.7",
	install_requires=[
		"colorama",
		"subpop >= 2.0.0",
		"Jinja2 >= 3",
		"PyYAML"
	],
	packages=setuptools.find_packages(),
	data_files=pkgr.generate_data_files(),
	cmdclass={"install": SubPopSetupInstall},
)

# vim: ts=4 sw=4 noet



#!/usr/bin/python3

import logging
import os
import pathlib
import shutil

from boxer.common import parse_args
from boxer.model import set_model
from boxer.pretty_logging import TornadoPrettyLogFormatter
from subpop.config import SubPopModel, ConfigurationError
from boxer.context import GitRepositoryLocator


class BoxerConfig(SubPopModel):
	"""
	This class contains configuration settings common to all the metatools plugins and tools.
	"""

	logger_name = "boxer"
	stage = None
	root = None
	tmp = None
	sub = None

	def __init__(self):
		super().__init__()

	async def initialize(self, stage=None, debug=False):

		if not stage:
			raise ValueError("stage not defined.")
		stage = pathlib.Path(stage).expanduser().resolve()
		if not stage.exists():
			raise FileNotFoundError(f"Cannot find stage: {stage}")
		self.stage = stage
		try:
			self.root = GitRepositoryLocator().root
		except ConfigurationError:
			self.root = f"/var/tmp/boxer_{os.getlogin()}_{os.getpid()}"
		self.tmp = self.root + "/tmp"
		if os.path.exists(self.tmp):
			shutil.rmtree(self.tmp)
		os.makedirs(self.tmp, exist_ok=False, mode=0o700)
		self.log = logging.getLogger(self.logger_name)
		self.log.propagate = False
		if debug:
			self.debug = debug
			self.log.setLevel(logging.DEBUG)
		else:
			self.log.setLevel(logging.INFO)
		channel = logging.StreamHandler()
		channel.setFormatter(TornadoPrettyLogFormatter())
		self.log.addHandler(channel)
		if debug:
			self.log.warning("DEBUG enabled")
		set_model("containers", self)


class DockerConfig(BoxerConfig):

	tag = None
	push = None
	sub = "docker"

	CLI_CONFIG = {
		"push": {"default": False, "action": "store_true"},
		"tag": {"default": None, "action": "store", "help": "Set a tag/name for the resultant container"},
	}

	async def initialize(self, stage=None, debug=False, extra_args=None):
		local_args = parse_args(self.CLI_CONFIG, extra_args)
		await super().initialize(stage=stage, debug=debug)
		self.push = local_args.push
		self.tag = local_args.tag


class SingularityConfig(BoxerConfig):

	out = None
	force = False
	sub = "singularity"

	CLI_CONFIG = {
		"out": {"default": None, "action": "store", "help": "Set an output filename/path for .sif file for singularity container."},
		"force": {"default": False, "action": "store_true", "help": "Force overwrite of existing singularity container."},
	}

	async def initialize(self, stage=None, debug=False, extra_args=None):
		local_args = parse_args(self.CLI_CONFIG, extra_args)
		await super().initialize(stage=stage, debug=debug)
		self.out = local_args.out
		self.force = local_args.force
